"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.UtxoCore = void 0;
const axios_1 = __importDefault(require("axios"));
const __1 = require("..");
const axios_rate_limit_1 = __importDefault(require("../axios-rate-limiter/axios-rate-limit"));
const LiteBlock_1 = require("../base-objects/blocks/LiteBlock");
const StatusBase_1 = require("../base-objects/StatusBase");
const genericMccTypes_1 = require("../types/genericMccTypes");
const constants_1 = require("../utils/constants");
const utils_1 = require("../utils/utils");
const utxoUtils_1 = require("../utils/utxoUtils");
const DEFAULT_TIMEOUT = 60000;
const DEFAULT_RATE_LIMIT_OPTIONS = {
    maxRPS: 5,
};
class UtxoCore {
    constructor(createConfig) {
        var _a;
        let client = axios_1.default.create({
            baseURL: createConfig.url,
            timeout: ((_a = createConfig.rateLimitOptions) === null || _a === void 0 ? void 0 : _a.timeoutMs) || DEFAULT_TIMEOUT,
            headers: { "Content-Type": "application/json" },
            auth: {
                username: createConfig.username,
                password: createConfig.password,
            },
            validateStatus: function (status) {
                return (status >= 200 && status < 300) || status == 500;
            },
        });
        this.client = (0, axios_rate_limit_1.default)(client, Object.assign(Object.assign({}, DEFAULT_RATE_LIMIT_OPTIONS), createConfig.rateLimitOptions));
        this.inRegTest = createConfig.inRegTest || false;
        this.loggingObject = createConfig.loggingOptions ? (0, utils_1.fillWithDefault)(createConfig.loggingOptions) : (0, utils_1.defaultMccLoggingObject)();
        // This has to be shadowed
        this.transactionConstructor = __1.UtxoTransaction;
        this.blockConstructor = __1.UtxoBlock;
        this.chainType = genericMccTypes_1.ChainType.BTC;
    }
    isHealthy() {
        return __awaiter(this, void 0, void 0, function* () {
            // WIP
            return true;
        });
    }
    ///////////////////////////////////////////////////////////////////////////////////////
    // Read part of RPC Client ////////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    /**
     * Get network info
     * @returns network info details
     */
    getNetworkInfo(retry = 0) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let res = yield this.client.post("", {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getnetworkinfo",
                    params: [],
                });
                // Error codes https://github.com/bitcoin/bitcoin/blob/master/src/rpc/protocol.h
                if ((0, utxoUtils_1.utxo_check_expect_empty)(res.data)) {
                    return null;
                }
                (0, utxoUtils_1.utxo_ensure_data)(res.data);
                return res.data;
            }
            catch (error) {
                if (retry === 0) {
                    this.loggingObject.exceptionCallback(error, `UtxCore::getNetworkInfo() (failed after max retries)`);
                    return null;
                }
                yield (0, utils_1.sleepMs)(100 + (0, utils_1.getSimpleRandom)(500));
                this.loggingObject.exceptionCallback(error, `UtxCore::getNetworkInfo() (retry ${retry})`);
                return yield this.getNetworkInfo(retry - 1);
            }
        });
    }
    /**
     * Get transaction details
     * @param txId Transaction id
     * @param options provide verbose:boolean, set true if you want more info such as block hash...
     * @returns transaction details
     */
    getTransaction(txId, options, retry = 3) {
        return __awaiter(this, void 0, void 0, function* () {
            if (constants_1.PREFIXED_STD_TXID_REGEX.test(txId)) {
                txId = (0, utils_1.unPrefix0x)(txId);
            }
            // TODO trow if txid does not match expected input
            try {
                let verbose = true; // by default getting transaction is in verbose mode
                let unTxId = (0, utils_1.unPrefix0x)(txId);
                let res = yield this.client.post("", {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getrawtransaction",
                    params: [unTxId, verbose],
                });
                // Error codes https://github.com/bitcoin/bitcoin/blob/master/src/rpc/protocol.h
                if ((0, utxoUtils_1.utxo_check_expect_empty)(res.data)) {
                    return null;
                }
                (0, utxoUtils_1.utxo_ensure_data)(res.data);
                return new this.transactionConstructor(res.data.result);
            }
            catch (error) {
                if (retry === 0) {
                    this.loggingObject.exceptionCallback(error, `UtxCore::getTransaction(${txId}) (failed after max retries)`);
                    return null;
                }
                yield (0, utils_1.sleepMs)(100 + (0, utils_1.getSimpleRandom)(500));
                this.loggingObject.exceptionCallback(error, `UtxCore::getTransaction(${txId}) (retry ${retry})`);
                return yield this.getTransaction(txId, options, retry - 1);
            }
        });
    }
    /**
     * Get Block height (number of blocks) from connected chain
     * @returns block height (block count)
     */
    getBlockHeight(retry = 3) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let res = yield this.client.post("", {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getblockcount",
                    params: [],
                });
                (0, utxoUtils_1.utxo_ensure_data)(res.data);
                return res.data.result;
            }
            catch (error) {
                if (retry === 0) {
                    this.loggingObject.exceptionCallback(error, `UtxCore::getTransaction() (failed after max retries)`);
                    return 0;
                }
                yield (0, utils_1.sleepMs)(100 + (0, utils_1.getSimpleRandom)(500));
                this.loggingObject.exceptionCallback(error, `UtxCore::getTransaction() (retry ${retry})`);
                return yield this.getBlockHeight(retry - 1);
            }
        });
    }
    // async RTRgetBlockHeight(): Promise<number> {
    //    const baseFunc = async () => {
    //       let res = await this.client.post("", {
    //          jsonrpc: "1.0",
    //          id: "rpc",
    //          method: "getblockcount",
    //          params: [],
    //       });
    //       utxo_ensure_data(res.data);
    //       return res.data.result;
    //    };
    //    // return await retryTimes(baseFunc, 3, this.loggingObject.exceptionCallback);
    //    return await retry("RTRgetBlockHeight", baseFunc)
    // }
    /**
     * Get header information about the block
     * @param blockHash
     * @returns
     */
    getBlockHeader(blockHashOrHeight, retry = 3) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let blockHash = null;
                if (typeof blockHashOrHeight === "string") {
                    blockHash = blockHashOrHeight;
                    if (constants_1.PREFIXED_STD_BLOCK_HASH_REGEX.test(blockHash)) {
                        blockHash = (0, utils_1.unPrefix0x)(blockHash);
                    }
                    // TODO match with some regex
                }
                else if (typeof blockHashOrHeight === "number") {
                    blockHash = yield this.getBlockHashFromHeight(blockHashOrHeight);
                }
                let res = yield this.client.post("", {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getblockheader",
                    params: [blockHash],
                });
                if ((0, utxoUtils_1.utxo_check_expect_empty)(res.data)) {
                    return null;
                }
                (0, utxoUtils_1.utxo_ensure_data)(res.data);
                return res.data.result;
            }
            catch (error) {
                if (retry === 0) {
                    this.loggingObject.exceptionCallback(error, `UtxCore::getBlockHeader(${blockHashOrHeight}) (failed after max retries)`);
                    return null;
                }
                yield (0, utils_1.sleepMs)(100 + (0, utils_1.getSimpleRandom)(500));
                this.loggingObject.exceptionCallback(error, `UtxCore::getBlockHeader(${blockHashOrHeight}) (retry ${retry})`);
                return yield this.getBlockHeader(blockHashOrHeight, retry - 1);
            }
        });
    }
    /**
     * Gets the block from main mining tip with provided height
     * @param blockNumber Block height
     * @returns Block hash
     */
    getBlockHashFromHeight(blockNumber) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post("", {
                jsonrpc: "1.0",
                id: "rpc",
                method: "getblockhash",
                params: [blockNumber],
            });
            if ((0, utxoUtils_1.utxo_check_expect_block_out_of_range)(res.data)) {
                return null;
            }
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    /**
     * Returns the block information
     * @param blockHashOrHeight Provide either block hash or height of the block
     * @returns All available block information
     */
    getBlock(blockHashOrHeight, retry = 3) {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let blockHash = null;
                if (typeof blockHashOrHeight === "string") {
                    blockHash = blockHashOrHeight;
                }
                else if (typeof blockHashOrHeight === "number") {
                    blockHash = yield this.getBlockHashFromHeight(blockHashOrHeight);
                    if (blockHash === null) {
                        return null;
                    }
                }
                let params = [blockHash, 2];
                if (this.chainType === genericMccTypes_1.ChainType.DOGE) {
                    params = [blockHash, true];
                }
                let res = yield this.client.post("", {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getblock",
                    params: params,
                });
                if ((0, utxoUtils_1.utxo_check_expect_block_out_of_range)(res.data)) {
                    return null;
                }
                (0, utxoUtils_1.utxo_ensure_data)(res.data);
                return new this.blockConstructor(res.data.result);
            }
            catch (error) {
                if (retry === 0) {
                    this.loggingObject.exceptionCallback(error, `UtxCore::getBlock(${blockHashOrHeight}) (failed after max retries)`);
                    return null;
                }
                yield (0, utils_1.sleepMs)(100 + (0, utils_1.getSimpleRandom)(500));
                this.loggingObject.exceptionCallback(error, `UtxCore::getBlock(${blockHashOrHeight}) (retry ${retry})`);
                return yield this.getBlock(blockHashOrHeight, retry - 1);
            }
        });
    }
    getBlockTipsHepler(height_gte, mainBranchProcess) {
        return __awaiter(this, void 0, void 0, function* () {
            // : Promise<BlockTip[]>{
            const tips = yield this.getTopBlocks({ height_gte: height_gte, all_blocks: true });
            let mainBranchHashes = [];
            const activeTip = tips.filter((a) => a.status === "active")[0];
            const ActiveTip = new LiteBlock_1.LiteBlock({ hash: activeTip.hash, number: activeTip.height });
            if (mainBranchProcess !== undefined) {
                mainBranchHashes = yield (0, utxoUtils_1.recursive_block_tip)(this, ActiveTip, mainBranchProcess);
            }
            const allTips = tips.map((UtxoTip) => {
                const tempTips = [];
                if (UtxoTip.all_block_hashes) {
                    for (let hashIndex = 0; hashIndex < UtxoTip.all_block_hashes.length; hashIndex++) {
                        tempTips.push(new LiteBlock_1.LiteBlock({ hash: UtxoTip.all_block_hashes[hashIndex], number: UtxoTip.height - hashIndex }));
                    }
                }
                return tempTips;
            });
            return allTips.reduce((acc, nev) => acc.concat(nev), []).concat(mainBranchHashes);
        });
    }
    /**
     * Get only tips of all the branches
     * @param height_gte
     * @returns
     */
    getBlockTips(height_gte) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.getBlockTipsHepler(height_gte);
        });
    }
    getTopLiteBlocks(branch_len) {
        return __awaiter(this, void 0, void 0, function* () {
            const height = yield this.getBlockHeight();
            return this.getBlockTipsHepler(height - branch_len, branch_len);
        });
    }
    /**
     * Get an array of all alternative chain tips
     * @returns
     */
    getTopBlocks(option) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(``, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "getchaintips",
                params: [],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            const gte_height = (option === null || option === void 0 ? void 0 : option.height_gte) || 0;
            let response = res.data.result.filter((el) => {
                return el.height >= gte_height;
            });
            if (option !== undefined) {
                if (option.all_blocks !== undefined) {
                    let extended = response.map((el) => (0, utxoUtils_1.recursive_block_hash)(this, el.hash, el.branchlen));
                    extended = yield Promise.all(extended);
                    for (let i = 0; i < response.length; i++) {
                        response[i].all_block_hashes = extended[i];
                    }
                }
            }
            return response;
        });
    }
    /**
     * TODO implement
     */
    getNodeStatus() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                let res = yield this.client.post(``, {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getnetworkinfo",
                    params: [],
                });
                (0, utxoUtils_1.utxo_ensure_data)(res.data);
                let bres = yield this.client.post(``, {
                    jsonrpc: "1.0",
                    id: "rpc",
                    method: "getblockchaininfo",
                    params: [],
                });
                (0, utxoUtils_1.utxo_ensure_data)(bres.data);
                return new StatusBase_1.UtxoNodeStatus(Object.assign(Object.assign({}, bres.data.result), res.data.result));
            }
            catch (e) {
                return null;
            }
        });
    }
    ///////////////////////////////////////////////////////////////////////////////////////
    // Write part of RPC Client ///////////////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    /**
     * Creates a new wallet on node's database
     * @param walletLabel label of your wallet used as a reference for future use
     * @returns name of the created wallet and possible warnings
     */
    createWallet(walletLabel) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post("", {
                jsonrpc: "1.0",
                id: "rpc",
                method: "createwallet",
                params: [walletLabel],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            // TODO try to import wallet if it already exists but is not imported
            return res.data.result;
        });
    }
    /**
     * loads the wallet if it exist on node, but it has to be relisted
     * @param walletLabel wallet label to load
     * @returns
     */
    loadWallet(walletLabel) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post("", {
                jsonrpc: "1.0",
                id: "rpc",
                method: "loadwallet",
                params: [walletLabel],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    /**
     *
     * @param walletLabel label of wallet, if you dont have one create one with createWallet
     * @param label label of address within wallet (default to "")
     * @param address_type type of address (default to "legacy") options = ["legacy", "p2sh-segwit", "bech32"]
     * @returns
     */
    createAddress(walletLabel, addressLabel = "", address_type = "legacy") {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "getnewaddress",
                params: [addressLabel, address_type],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    /**
     * List all wallets on node
     * @returns
     */
    listAllWallets() {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(``, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "listwallets",
                params: [],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    /**
     * List all addresses by their label on provided wallet
     * @param walletLabel label of the parent wallet we want to list addresses
     * @param addressLabel label of the addresses we want to list
     * @returns
     */
    listAllAddressesByLabel(walletLabel, addressLabel = "") {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "getaddressesbylabel",
                params: [addressLabel],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            let address_labels = Object.keys(res.data.result);
            let response_array = [];
            for (let addL of address_labels) {
                response_array.push({ address: addL, purpose: res.data.result[addL].purpose });
            }
            return response_array;
        });
    }
    /**
     * List all unspent transactions that happened between min and max blocks before current block
     * If we are in block 100 and set min to 10 and max to 40 we will get all transactions that happened
     * between block 60 and 90
     * @param walletLabel
     * @param min min block offset
     * @param max max block offset
     * @returns
     */
    listUnspentTransactions(walletLabel, min = 0, max = 1e6) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "listunspent",
                params: [min, max],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    createRawTransaction(walletLabel, vin, out) {
        return __awaiter(this, void 0, void 0, function* () {
            let voutArr = "[";
            let first = true;
            for (let i of out) {
                if (first) {
                    first = false;
                }
                else {
                    voutArr += ",";
                }
                let row = `{"${i.address}" : ${i.amount}}`;
                voutArr += row;
            }
            voutArr += "]";
            let VoutArr = JSON.parse(voutArr);
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "createrawtransaction",
                params: [vin, VoutArr],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    signRawTransaction(walletLabel, rawTx, keysList) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "signrawtransactionwithkey",
                params: [rawTx, keysList],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    /**
     * Send raw transaction
     * @param walletLabel the label of the wallet we are sending the transaction from
     * @param signedRawTx hash of signed transaction
     * @returns transaction sending status
     */
    sendRawTransaction(walletLabel, signedRawTx) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "sendrawtransaction",
                params: [signedRawTx],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    /**
     * Send raw transaction and wait for it to be in next block
     * @param walletLabel the label of the wallet we are sending the transaction from
     * @param signedRawTx hash of signed transaction
     * @returns transaction sending status
     */
    sendRawTransactionInBlock(walletLabel, signedRawTx) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "sendrawtransaction",
                params: [signedRawTx],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            let tx = yield this.getTransaction(res.data.result);
            while (tx === null || tx === void 0 ? void 0 : tx.data.blockhash) {
                yield (0, utils_1.sleepMs)(3000);
                tx = yield this.getTransaction(res.data.result);
            }
            return res.data.result;
        });
    }
    /**
     * Get private key from wallet
     * @notice Dont share this with anyone
     * @param walletLabel wallet label that owns the address
     * @param address
     * @returns private key
     */
    getPrivateKey(walletLabel, address) {
        return __awaiter(this, void 0, void 0, function* () {
            let res = yield this.client.post(`wallet/${walletLabel}`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "dumpprivkey",
                params: [address],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
    ///////////////////////////////////////////////////////////////////////////////////////
    // Regtest faucet part of RPC Client //////////////////////////////////////////////////
    ///////////////////////////////////////////////////////////////////////////////////////
    /**
     *
     * @dev Note that miner has to be registered in miner wallet
     * @dev only in regtest mode
     * @param address
     * @param amount
     * @returns
     */
    fundAddress(address, amount) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.inRegTest) {
                throw Error("You have to run client in regression test mode to use this ");
            }
            let res = yield this.client.post(`wallet/miner`, {
                jsonrpc: "1.0",
                id: "rpc",
                method: "sendtoaddress",
                params: [address, amount],
            });
            (0, utxoUtils_1.utxo_ensure_data)(res.data);
            return res.data.result;
        });
    }
}
exports.UtxoCore = UtxoCore;
//# sourceMappingURL=UtxoCore.js.map